import os
import subprocess
import tempfile
from pathlib import Path

import pytest


@pytest.fixture
def tempdir_path():
    with tempfile.TemporaryDirectory() as tempdir_path:
        yield Path(tempdir_path).resolve()


@pytest.fixture
def init_git_and_commit(call_git):
    def _init_git_and_commit(path: Path, with_remote=True) -> None:
        call_git("git init", cwd=path)
        call_git("git add --all .", cwd=path)
        call_git('git commit -m "initial"', cwd=path)
        if with_remote:
            call_git(
                "git remote add origin https://gitlab.localhost/nonexisting/repo.git",
                cwd=path,
            )

    return _init_git_and_commit


@pytest.fixture
def call_git():
    def _call_git(shell_cmd: str, cwd: Path) -> None:
        env = {
            "USERNAME": "tests_scaraplate",
            "EMAIL": "pytest@scaraplate",
            "PATH": os.getenv("PATH", os.defpath),
        }
        subprocess.run(shell_cmd, shell=True, check=True, cwd=cwd, env=env, timeout=5)

    return _call_git



import click

from scaraplate.rollup import rollup as _rollup


@click.group()
@click.version_option()
def main():
    pass


@main.command()
@click.argument("template_dir", type=click.Path(exists=True, file_okay=False))
@click.argument("target_project_dir", type=click.Path(file_okay=False))
@click.option(
    "--no-input",
    is_flag=True,
    help=(
        "Do not prompt for missing data. Cookiecutter will use "
        "the defaults provided by the `cookiecutter.json` "
        "in the TEMPLATE_DIR."
    ),
)
def rollup(template_dir: str, target_project_dir: str, no_input: bool) -> None:
    """Rollup (apply) the cookiecutter template.

    The template from TEMPLATE_DIR is applied on top of TARGET_PROJECT_DIR.

    TEMPLATE_DIR must be a local path to the location of a git repo with
    the cookiecutter template to apply. It must contain `scaraplate.yaml`
    file in the root. The TEMPLATE_DIR must be in a git repo, because
    some strategies require the commit hash at HEAD and the git remote URL.

    TARGET_PROJECT_DIR should point to the directory where the template
    should be applied. Might not exist -- it will be created then.
    For monorepos this should point to the subproject inside the monorepo.
    """
    _rollup(
        template_dir=template_dir,
        target_project_dir=target_project_dir,
        no_input=no_input,
    )


if __name__ == "__main__":
    main(prog_name="scaraplate")  # pylint: disable=unexpected-keyword-arg



import contextlib
import json
import os
from pathlib import Path
from typing import Dict, Optional
from unittest.mock import sentinel

import pytest

from scaraplate.config import ScaraplateYaml
from scaraplate.cookiecutter import ScaraplateConf
from scaraplate.rollup import (
    get_project_dest,
    get_strategy,
    get_target_project_cookiecutter_context,
    get_template_root_and_dir,
    rollup,
)


@contextlib.contextmanager
def with_working_directory(target_dir: Path):
    cur = os.getcwd()
    os.chdir(target_dir)
    try:
        yield
    finally:
        os.chdir(cur)


@pytest.mark.parametrize("apply_count", [1, 2])
def test_rollup_fuzzy(tempdir_path, apply_count, init_git_and_commit):
    template_path = tempdir_path / "template"
    target_project_path = tempdir_path / "test"

    # Prepare template
    cookiecutter_path = template_path / "{{cookiecutter.project_dest}}"
    cookiecutter_path.mkdir(parents=True)
    (cookiecutter_path / "README.md").write_text(
        "{{ cookiecutter.project_dest }} mock!"
    )
    (cookiecutter_path / "setup.py").write_text("#!/usr/bin/env python\n")
    (cookiecutter_path / "setup.py").chmod(0o755)
    (cookiecutter_path / "sense_vars").write_text("{{ cookiecutter|jsonify }}\n")
    (template_path / "cookiecutter.json").write_text('{"project_dest": "test"}')
    (template_path / "scaraplate.yaml").write_text(
        """
default_strategy: scaraplate.strategies.Overwrite
strategies_mapping:
  setup.py: scaraplate.strategies.TemplateHash
        """
    )
    init_git_and_commit(template_path)

    # Apply template (possibly multiple times)
    for i in range(apply_count):
        rollup(
            template_dir=str(template_path),
            target_project_dir=str(target_project_path),
            no_input=True,
        )

        assert "test mock!" == (target_project_path / "README.md").read_text()
        assert 0o755 == (0o777 & (target_project_path / "setup.py").stat().st_mode)

        with open((target_project_path / "sense_vars"), "rt") as f:
            assert json.load(f) == {
                # fmt: off
                "_template": "template",
                "project_dest": "test",
                # fmt: on
            }


def test_get_project_dest(tempdir_path: Path) -> None:
    target = tempdir_path / "myproject"
    with with_working_directory(tempdir_path):
        assert "myproject" == get_project_dest(Path("myproject"))

    target.mkdir()
    with with_working_directory(target):
        assert "myproject" == get_project_dest(Path("."))


def test_get_template_root_and_dir(tempdir_path: Path) -> None:
    target = tempdir_path / "myproject"
    target.mkdir()

    with with_working_directory(tempdir_path):
        assert (tempdir_path, "myproject") == get_template_root_and_dir(
            Path("myproject")
        )
        assert (tempdir_path, "myproject") == get_template_root_and_dir(
            tempdir_path / "myproject"
        )

    with with_working_directory(target):
        assert (tempdir_path, "myproject") == get_template_root_and_dir(Path("."))


@pytest.mark.parametrize(
    "contents, expected_context",
    [
        (None, {}),
        ("", {}),
        (
            """
[cookiecutter_context]
metadata_author = Usermodel @ Rambler&Co
coverage_fail_under = 90
project_monorepo_name =
""",
            {
                "metadata_author": "Usermodel @ Rambler&Co",
                "coverage_fail_under": "90",
                "project_monorepo_name": "",
            },
        ),
    ],
)
def test_get_target_project_cookiecutter_context(
    tempdir_path: Path, contents: Optional[str], expected_context: Dict
) -> None:
    if contents is not None:
        (tempdir_path / ".scaraplate.conf").write_text(contents)

    scaraplate_yaml = ScaraplateYaml(
        default_strategy=sentinel.default,
        strategies_mapping={},
        git_remote_type=None,
        cookiecutter_context_type=ScaraplateConf,
    )

    assert expected_context == get_target_project_cookiecutter_context(
        tempdir_path, scaraplate_yaml
    )


def test_get_strategy():
    scaraplate_yaml = ScaraplateYaml(
        default_strategy=sentinel.default,
        strategies_mapping={
            "Jenkinsfile": sentinel.jenkinsfile,
            "some/nested/setup.py": sentinel.nested_setup_py,
            "src/*/__init__.py": sentinel.glob_init,
        },
        git_remote_type=None,
        cookiecutter_context_type=ScaraplateConf,
    )

    assert sentinel.default is get_strategy(scaraplate_yaml, Path("readme"))
    assert sentinel.jenkinsfile is get_strategy(scaraplate_yaml, Path("Jenkinsfile"))
    assert sentinel.nested_setup_py is get_strategy(
        scaraplate_yaml, Path("some/nested/setup.py")
    )
    assert sentinel.glob_init is get_strategy(
        scaraplate_yaml, Path("src/my_project/__init__.py")
    )



from click.testing import CliRunner

from scaraplate.__main__ import main


def test_help():
    runner = CliRunner()
    result = runner.invoke(main, ["--help"])
    assert result.exit_code == 0


def test_rollup_help():
    runner = CliRunner()
    result = runner.invoke(main, ["rollup", "--help"])
    assert result.exit_code == 0
